# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Dict, Tuple

import requests
from adr.util.memoize import memoize, memoized_property

from mozci.errors import PushNotFound
from mozci.util.req import get_session


class HGMO:
    # urls
    BASE_URL = "https://hg.mozilla.org/"
    AUTOMATION_RELEVANCE_TEMPLATE = BASE_URL + "{branch}/json-automationrelevance/{rev}"
    JSON_TEMPLATE = BASE_URL + "{branch}/rev/{rev}?style=json"
    JSON_PUSHES_TEMPLATE = (
        BASE_URL
        + "{branch}/json-pushes?version=2&startID={push_id_start}&endID={push_id_end}"
    )

    # instance cache
    CACHE: Dict[Tuple[str, str], HGMO] = {}

    def __init__(self, rev, branch="autoland"):
        self.context = {
            "branch": "integration/autoland" if branch == "autoland" else branch,
            "rev": rev,
        }

    @staticmethod
    def create(rev, branch="autoland"):
        key = (branch, rev[:12])
        if key in HGMO.CACHE:
            return HGMO.CACHE[key]
        instance = HGMO(rev, branch)
        HGMO.CACHE[key] = instance
        return instance

    def _get_resource(self, url):
        try:
            r = get_session().get(url)
        except requests.exceptions.RetryError as e:
            raise PushNotFound(f"{e} error when getting {url}", **self.context)

        if r.status_code == 404:
            raise PushNotFound(f"{r.status_code} response from {url}", **self.context)

        r.raise_for_status()
        return r.json()

    @memoized_property
    def changesets(self):
        url = self.AUTOMATION_RELEVANCE_TEMPLATE.format(**self.context)
        return self._get_resource(url)["changesets"]

    @memoized_property
    def data(self):
        url = self.JSON_TEMPLATE.format(**self.context)
        return self._get_resource(url)

    def __getitem__(self, k):
        return self.data[k]

    def get(self, k, default=None):
        return self.data.get(k, default)

    @memoize
    def json_pushes(self, push_id_start, push_id_end):
        url = self.JSON_PUSHES_TEMPLATE.format(
            push_id_start=push_id_start,
            push_id_end=push_id_end,
            **self.context,
        )
        return self._get_resource(url)["pushes"]

    @property
    def pushid(self):
        return self.changesets[0]["pushid"]

    @property
    def pushhead(self):
        return self.changesets[0]["pushhead"]

    @property
    def backouts(self):
        # Sometimes json-automationrelevance doesn't return all commits of a push.
        # https://bugzilla.mozilla.org/show_bug.cgi?id=1641729
        if self.pushhead not in {changeset["node"] for changeset in self.changesets}:
            return HGMO.create(self.pushhead, branch=self.context["branch"]).backouts

        return {
            changeset["node"]: [node["node"] for node in changeset["backsoutnodes"]]
            for changeset in self.changesets
            if len(changeset["backsoutnodes"])
        }

    @property
    def bugs(self):
        return set(
            bug["no"] for changeset in self.changesets for bug in changeset["bugs"]
        )

    @property
    def bugs_without_backouts(self):
        return {
            bug["no"]: changeset["node"]
            for changeset in self.changesets
            for bug in changeset["bugs"]
            if len(changeset["backsoutnodes"]) == 0
        }



# -*- coding: utf-8 -*-

from collections import defaultdict

from loguru import logger

from mozci.data.base import DataSource

try:
    from treeherder.model.models import Job
except ImportError:
    Job = None


class TreeherderSource(DataSource):
    """Uses ORM to query Treeherder."""

    name = "treeherder"
    supported_contracts = (
        "push_tasks",
        "push_tasks_tags",
        "push_test_groups",
    )

    @classmethod
    def normalize(cls, jobs):
        retries = defaultdict(int)
        items = []

        for job in jobs:
            task_id = job.taskcluster_metadata.task_id
            retry_id = job.taskcluster_metadata.retry_id

            # If a task is re-run, use the data from the last run.
            if retry_id < retries[task_id]:
                logger.trace(f"Skipping {job} because there is a newer run of it.")
                continue

            retries[task_id] = retry_id

            note = ""
            notes = job.jobnote_set.all()
            if len(notes):
                note = str(notes[0].text)

            result = {
                "id": str(task_id),
                "label": job.job_type.name,
                "result": job.result,
                "classification": job.failure_classification.name,
                "classification_note": note,
                "duration": Job.get_duration(
                    job.submit_time, job.start_time, job.end_time
                ),
            }
            items.append(result)
        return items

    def run_push_tasks(self, **kwargs):
        if Job:
            jobs = (
                Job.objects.filter(
                    push__revision=kwargs["rev"], repository__name=kwargs["branch"]
                )
                .exclude(
                    tier=3,
                    result="retry",
                    job_type__name="Gecko Decision Task",
                    job_type__name__startswith="Action",
                )
                .select_related(
                    "push",
                    "job_type",
                    "taskcluster_metadata",
                    "failure_classification",
                    "repository",
                    "repository__repository_group",
                )
                .prefetch_related("jobnote_set")
            )

            tasks = []

            for task in self.normalize(jobs):
                if task.get("tags"):
                    task["tags"] = {t["name"]: t["value"] for t in task["tags"]}

                tasks.append(task)
            return tasks

        logger.trace(
            "Unable to reach Treeherder as a datasource because the Job model "
            "was not available to import."
        )
        return []

    def run_push_tasks_tags(self, **kwargs):
        # These are not needed for Push Health
        return {}

    def run_push_test_groups(self, **kwargs):
        # These are not needed for Push Health
        return {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mozci',
 'mozci.data',
 'mozci.data.sources',
 'mozci.data.sources.activedata',
 'mozci.data.sources.treeherder',
 'mozci.util']

package_data = \
{'': ['*'],
 'mozci': ['queries/*'],
 'mozci.data.sources.activedata': ['queries/*']}

install_requires = \
['adr>=0.19,<0.20',
 'appdirs>=1.4.4,<2.0.0',
 'cachy>=0.3.0,<0.4.0',
 'flake8>=3.8.3,<4.0.0',
 'loguru>=0.5.1,<0.6.0',
 'requests>=2.24.0,<3.0.0',
 'taskcluster_urls>=12.1,<14.0',
 'tomlkit>=0.6.0,<0.7.0',
 'voluptuous>=0.11.7,<0.12.0']

setup_kwargs = {
    'name': 'mozci',
    'version': '0.4.0',
    'description': '',
    'long_description': None,
    'author': 'Andrew Halberstadt',
    'author_email': 'ahal@mozilla.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import functools

import requests
import taskcluster_urls as liburls

from mozci.util import yaml
from mozci.util.req import get_session

PRODUCTION_TASKCLUSTER_ROOT_URL = "https://firefox-ci-tc.services.mozilla.com"


def _do_request(url, force_get=False, **kwargs):
    session = get_session()
    if kwargs and not force_get:
        response = session.post(url, **kwargs)
    else:
        response = session.get(url, stream=True, **kwargs)
    if response.status_code >= 400:
        # Consume content before raise_for_status, so that the connection can be
        # reused.
        response.content
    response.raise_for_status()
    return response


def _handle_artifact(path, response):
    if path.endswith(".json"):
        return response.json()
    if path.endswith(".yml"):
        return yaml.load_stream(response.text)
    response.raw.read = functools.partial(response.raw.read, decode_content=True)
    return response.raw


def get_artifact_url(task_id, path, old_deployment=False):
    if not old_deployment:
        return liburls.api(
            PRODUCTION_TASKCLUSTER_ROOT_URL,
            "queue",
            "v1",
            f"task/{task_id}/artifacts/{path}",
        )
    else:
        return f"https://queue.taskcluster.net/v1/task/{task_id}/artifacts/{path}"


def get_artifact(task_id, path):
    """
    Returns the artifact with the given path for the given task id.

    If the path ends with ".json" or ".yml", the content is deserialized as,
    respectively, json or yaml, and the corresponding python data (usually
    dict) is returned.
    For other types of content, a file-like object is returned.
    """
    try:
        response = _do_request(get_artifact_url(task_id, path))
    except requests.exceptions.HTTPError as e:
        if e.response.status_code != 404:
            raise

        response = _do_request(get_artifact_url(task_id, path, old_deployment=True))

    return _handle_artifact(path, response)


def list_artifacts(task_id):
    try:
        response = _do_request(get_artifact_url(task_id, "").rstrip("/"))
    except requests.exceptions.HTTPError:
        response = _do_request(
            get_artifact_url(task_id, "", old_deployment=True).rstrip("/")
        )
    return response.json()["artifacts"]


def get_index_url(index_path):
    return liburls.api(
        PRODUCTION_TASKCLUSTER_ROOT_URL, "index", "v1", f"task/{index_path}"
    )


def find_task_id(index_path, use_proxy=False):
    response = _do_request(get_index_url(index_path))
    return response.json()["taskId"]
